package descartes.kieker4dql.dqlconnector;

import java.util.ArrayList;

import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;

import edu.kit.ipd.descartes.core.NamedElement;
import edu.kit.ipd.descartes.identifier.Identifier;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ExternalCallAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ExternalCallFrequency;
import edu.kit.ipd.descartes.mm.containerrepository.ContainerTemplate;
import edu.kit.ipd.descartes.mm.resourceconfiguration.ProcessingResourceSpecification;
import edu.kit.ipd.descartes.mm.resourcelandscape.Container;
import edu.kit.ipd.descartes.mm.usageprofile.SystemCallUserAction;
import edu.kit.ipd.descartes.mm.usageprofile.UsageScenario;
import edu.kit.ipd.descartes.mm.usageprofile.WorkloadType;
import edu.kit.ipd.descartes.mm.usageprofile.impl.ClosedWorkloadTypeImpl;
import edu.kit.ipd.descartes.mm.usageprofile.impl.OpenWorkloadTypeImpl;
import edu.kit.ipd.descartes.solving.DMLModels;
import edu.kit.ipd.descartes.solving.ModelValidator;
import tools.descartes.dql.connector.AbstractOSGiQueryConnector;
import tools.descartes.dql.connector.ModelStructureQueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;
import tools.descartes.dql.core.engine.util.DQLLogger;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.Resource;
import tools.descartes.dql.models.mapping.domain.Service;
import tools.descartes.dql.models.mapping.domain.impl.DomainFactoryImpl;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

public class ModelStructureQueryConnectorImpl extends
		AbstractOSGiQueryConnector implements ModelStructureQueryConnector {
	private final Logger log = DQLLogger.getLogger(this.getClass().getName());
	private final DomainFactory domainFactory = DomainFactoryImpl.init();

	public ModelStructureQueryConnectorImpl() {
		super();
	}

	/**
	 * Creates and attaches probes to a list of probes
	 * 
	 * @param metrics
	 *            Metrics to be attached as probes
	 * @param probes
	 *            List of probes, where probes for metrics shall be attached
	 */
	private void attachProbes(final String[] metrics, final EList<Probe> probes) {
		Probe probe = null;

		for (final String metric : metrics) {
			probe = domainFactory.createProbe();
			probe.setMetricName(metric);

			probes.add(probe);
		}
	}

	@Override
	public String getBundleProperty(final String key) {
		return Activator.getContext().getProperty(key);
	}

	/**
	 * Processes a found DML model identifier and performs individual
	 * processing depending on the type of entity found. The model entity found
	 * is stored in a suitable container (resources or services).
	 * 
	 * @param resources
	 *            The destination for a resource found
	 * @param services
	 *            The destination for a service found
	 * @param result
	 *            The model entity to be analyzed	 
	 */
	private void processIdentifier(
			final EList<tools.descartes.dql.models.mapping.domain.Resource> resources,
			final EList<Service> services, final EObject result) {
		final Identifier identifier = (Identifier) result;
		final NamedElement namedElement = (NamedElement) result;

		if (result instanceof SystemCallUserAction
				|| result instanceof ExternalCallAction
				|| result instanceof ExternalCallFrequency
				|| result instanceof UsageScenario) {
			// Found a Service
			Entity entity = domainFactory.createService();
			entity.setIdentifier(identifier.getId());
			entity.setAlias(namedElement.getName());
			services.add((Service) entity);
		} else if (result instanceof ProcessingResourceSpecification) {
			// Found a Resource, apply workaround
			processProcessingResourceSpec(resources, (ProcessingResourceSpecification) result);
		} else {
			log.error("Well, this is a bug... An explicitly requested entity is not processed. Skipping.");
			return;
		}

	}

	/**
	 * Attaches available probes to resources
	 * 
	 * @param domainFactory
	 * @param resources
	 */
	private void processProbesForResources(final EList<Resource> resources) {
		// TODO support DMLHelper.METRIC_THROUGHPUT also for resources
		final String[] metrics = new String[] { DMLHelper.METRIC_UTILIZATION };

		for (final Resource r : resources) {
			attachProbes(metrics, r.getProbes());
		}
	}

	/**
	 * Attaches available probes to services
	 * 
	 * @param domainFactory
	 * @param resources
	 */
	private void processProbesForServices(final EList<Service> services) {
		final String[] metrics = new String[] { DMLHelper.METRIC_RESPONSE_TIME_AVG, DMLHelper.METRIC_THROUGHPUT };

		for (final Service s : services) {
			attachProbes(metrics, s.getProbes());
		}
	}

	/**
	 * Process a ProcessingResourceSpecification and add it to the EntityMapping list of
	 * resources, the alias is enriched by a specific prefix in order to understand where the 
	 * ProcessingResource comes from.
	 * 
	 * @param resources
	 *            The EntityMapping list of resources
	 * @param identifier
	 *            The DML Identifier of the resource to be modified
	 * @param namedElement
	 *            The DML NamedElement of the resource to be modified
	 * @return The created EntityMapping representation 
	 */
	private void processProcessingResourceSpec(
			final EList<tools.descartes.dql.models.mapping.domain.Resource> resources,
			final ProcessingResourceSpecification prs) {
		// compute understandable alias as concatenation of Strings of names of 
		// Container -> ProcessingResourceSpecification
		String alias = null;
		EObject parent = prs.eContainer();
		if (parent instanceof Container) {
			alias = ((Container) parent).getName() + "." + prs.getName();
		} else if (parent instanceof ContainerTemplate) {
			log.error("ProcessingResourceSpecification " + prs.getId() 
					+ " is not supported because ContainerTemplates are not yet supported.");
		} else {
			log.error("ProcessingResourceSpecification " + prs.getId() 
					+ " is not supported because its parent types cannot be resolved.");
		}
		
		if (null != alias) {
			Entity entity = domainFactory.createResource();
			entity.setIdentifier(prs.getId());
			entity.setAlias(alias);
			resources.add((Resource) entity);
		}
	}

	@Override
	public EntityMapping requestDoF(final EntityMapping request) {
		final EntityMapping response = EcoreUtil.copy(request);
		final String modelLocation = request.getModelLocation();
		final ArrayList<EObject> queryResults = new ArrayList<EObject>();
		final DMLModelFiles dmlModelFiles = DMLHelper.createDMLModelFiles(modelLocation);

		// Load and validate Model
		DMLModels models = DMLHelper.loadModels(dmlModelFiles);
		ModelValidator validator = new ModelValidator(models);
		if (!validator.isSetupValid()) {
			DMLHelper.openValidationErrorWindow(validator.getErrors());
			return null;
		}
		for (UsageScenario scenario : models.getUsageProfileModel().getUsageScenarios()) {
			WorkloadType workloadType = scenario.getWorkloadType();
			// "self.oclIsKindOf(ProcessingResourceSpecification)"
			if (workloadType.getClass().equals(OpenWorkloadTypeImpl.class)) {
				// queryResults.add(workloadType);
				queryResults.add(scenario);
			} else if (workloadType.getClass().equals(ClosedWorkloadTypeImpl.class)) {
				queryResults.add(scenario);
			}else{
				log.error("no workload specified " + workloadType);
			}
		}
		EList<DoF> dofs = response.getDoF();

		for (final EObject result : queryResults) {
			if (result instanceof Identifier) {
				final Identifier identifier = (Identifier) result;
				final NamedElement namedElement = (NamedElement) result;
				Entity entity = domainFactory.createDoF();
				entity.setIdentifier(identifier.getId());
				entity.setAlias(namedElement.getName());
				dofs.add((DoF) entity);
			} else {
				log.warn("Query returned unknown result type: " + result);
			}
		}

		return response;
	}

	@Override
	public EntityMapping requestEntities(final EntityMapping request) {
		final EntityMapping response = EcoreUtil.copy(request);
		final String modelLocation = request.getModelLocation();
		final ArrayList<EObject> queryResults = new ArrayList<EObject>();
		final DMLModelFiles dmlModelFiles = DMLHelper.createDMLModelFiles(modelLocation);

		//Load and validate Model
		DMLModels models = DMLHelper.loadModels(dmlModelFiles);
		ModelValidator validator = new ModelValidator(models);
		if (!validator.isSetupValid()){
			DMLHelper.openValidationErrorWindow(validator.getErrors());
			return null;
		}
		
		// When entities are requested and no alias is set, always supply
		// an alias if it is computable, e.g. the DML alias

		// Collect SystemCallUserActions
		// -> services [USAGEPROFILE]
		queryResults.addAll(DMLHelper.executeOCLQueryWithoutContext(
				new EObject[] { models.getUsageProfileModel() },
				DMLHelper.OCL_QUERY_SERVICES_ELSC));
		
		// Collect ExternalCallFrequencies
		// -> services [REPOSITORY]
		queryResults.addAll(DMLHelper.executeOCLQueryWithoutContext(
				new EObject[] { models.getResourceLandscapeModel() },
				DMLHelper.OCL_QUERY_SERVICES_ECFR));

		// Collect ExternalCallActions
		// -> services [REPOSITORY]
		queryResults.addAll(DMLHelper.executeOCLQueryWithoutContext(
				models.getRepositories().toArray(new EObject[models.getRepositories().size()]),
				DMLHelper.OCL_QUERY_SERVICES_ECAC));

		// Collect
		// ProcessingResourceSpecification
		// TODO CommunicationLinkResourceSpecification
		// -> resources [RESOURCELANDSCAPE]
		queryResults.addAll(DMLHelper.executeOCLQueryWithoutContext(
				new EObject[] { models.getResourceLandscapeModel() },
				DMLHelper.OCL_QUERY_RESOURCES));

		final EList<tools.descartes.dql.models.mapping.domain.Resource> resources = response.getResources();
		final EList<Service> services = response.getServices();

		for (final EObject result : queryResults) {
			if (result instanceof Identifier) {
				processIdentifier(resources, services, result);
			} else {
				log.warn("Query returned unknown result type: " + result);
			}
		}

		return response;
	}

	@Override
	public EntityMapping requestProbes(final EntityMapping request) {
		//Load and validate Model
		String modelLocation = request.getModelLocation();
		DMLModelFiles dmlModelFiles = DMLHelper.createDMLModelFiles(modelLocation);
		DMLModels models = DMLHelper.loadModels(dmlModelFiles);
		ModelValidator validator = new ModelValidator(models);
		if (!validator.isSetupValid()){
			DMLHelper.openValidationErrorWindow(validator.getErrors());
			return null;
		}
		
		final EntityMapping response = EcoreUtil.copy(request);
		processProbesForServices(response.getServices());
		processProbesForResources(response.getResources());
		return response;
	}

	@Override
	public boolean supportsDoFRequests() {
		return true;
	}

	@Override
	public boolean supportsEntityRequests() {
		return true;
	}

	@Override
	public boolean supportsProbeRequests() {
		return true;
	}

	@Override
	public void addQueryProgressListener(QueryProgressListener arg0) {
		// TODO Auto-generated method stub
		
	}

}
