package descartes.kieker4dql.dqlconnector.structures;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import descartes.kieker4dql.dqlconnector.structures.identifier.ComponentIdentifier;
import descartes.kieker4dql.dqlconnector.structures.identifier.ResourceIdentifier;
import descartes.kieker4dql.dqlconnector.structures.records.AbstractRecord;
import descartes.kieker4dql.dqlconnector.structures.records.EventRecord;
import descartes.kieker4dql.dqlconnector.structures.records.ResourceRecord;

/**
 * Stores list of event records and offers multiple analysis methods.
 * 
 * @author J�rgen Walter
 */
public class Trace {

	/** Comparator to sort EventRecord */
	@SuppressWarnings("rawtypes")
	private class ChronologicalEventRecordComparator implements	Comparator<AbstractRecord> {
		@Override
		public int compare(AbstractRecord event1, AbstractRecord event2) {
			if (event1.getTimestamp() < event2.getTimestamp()) {
				return -1;
			} else if (event1.getTimestamp() > event2.getTimestamp()) {
				return 1;
			} else {
				return 0;
			}
		}
	}
	
	//preset identifiers connected to Components/Resources (Components/Resources initialized via name only)
	private Map<ComponentIdentifier, Component> componentIdentifierToComponentMap; //must be preset
	private Map<ResourceIdentifier, Resource> resourceIdentifierToResourceMap; //must be preset

	// must be filled via input records => addEvent(...), addRessource(...)
	private List<EventRecord> events; //user input
	private List<ResourceRecord> resources; //user input

	// generated by invocation of "initializeMaps()" selecting valid records
	private Map<Component, Trace> componentToTraceMap;
	private Map<Resource, Trace> resourceToTraceMap;
	private Map<String, Request> idToRequestMap; //components' SessionID 

	// EventRecords and ResourceRecords that could not be mapped to a component 
	private List<EventRecord> undefinedEventRecords;
	private List<ResourceRecord> undefinedResourceRecords;

	/**
	 * 
	 * @return all incoming raw unfiltered resources
	 */
	public List<EventRecord> getEvents() {
		return events;
	}
	
	/**
	 * 
	 * @return all incoming raw unfiltered resources
	 */
	public List<ResourceRecord> getResources() {
		return resources;
	}

	public Map<ComponentIdentifier, Component> getComponentIdentifierToComponentMap() {
		return componentIdentifierToComponentMap;
	}

	public Map<ResourceIdentifier, Resource> getResourceIdentifierToResourceMap() {
		return resourceIdentifierToResourceMap;
	}

	public Map<Component, Trace> getComponentToTraceMap() {
		return componentToTraceMap;
	}

	public Map<Resource, Trace> getResourceToTraceMap() {
		return resourceToTraceMap;
	}
	
	public Map<String, Request> getIdToRequestMap() {
		return idToRequestMap;
	}

	public Trace() {
		this.events = new ArrayList<EventRecord>();
		this.undefinedEventRecords = new ArrayList<EventRecord>();
		this.resources = new ArrayList<ResourceRecord>();
		this.undefinedResourceRecords = new ArrayList<ResourceRecord>();
	}

	public void addEvent(EventRecord event) {
		events.add(event);
	}

	public void addResource(ResourceRecord event) {
		resources.add(event);
	}

	public Trace initializeMaps() {
		idToRequestMap = new HashMap<String, Request>();
		componentToTraceMap = new HashMap<Component, Trace>();
		resourceToTraceMap = new HashMap<Resource, Trace>();
		for (EventRecord event : events) {
			if (addToComponentMap(event)) {
				addToRequestMap(event);
			}
		}
		for (ResourceRecord resource : resources) {
			addToResourceMap(resource);
		}
		return this;
	}

	public void setComponentIdentifierToComponentMap(
			Map<ComponentIdentifier, Component> componentIdentifierToComponentMap) {
		this.componentIdentifierToComponentMap = componentIdentifierToComponentMap;
	}

	public void setResourceIdentifierToResourceMap(
			Map<ResourceIdentifier, Resource> resourceIdentifierToResourceMap) {
		this.resourceIdentifierToResourceMap = resourceIdentifierToResourceMap;
	}

	void sortChronological() {
		Collections.sort(events, new ChronologicalEventRecordComparator());
		Collections.sort(resources, new ChronologicalEventRecordComparator());
	}

	@Override
	public String toString() {
		StringBuffer sb = new StringBuffer();
		for (EventRecord event : events) {
			sb.append("\n\t" + event + "\n");
		}
		for (ResourceRecord record : resources) {
			sb.append("\n\t" + record + "\n");
		}
		return sb.toString();
	}

	private boolean addToComponentMap(EventRecord event) {
		ComponentIdentifier componentIdentifier = event.getIdentifier();
		Component component = componentIdentifierToComponentMap.get(componentIdentifier);

		//check if events matches a valid component, else it is stored as "undefined"
		if (component == null) {
			log.debug("componentIdentifier of \"" + event	+ "\" could not be mapped to a component \n");
			undefinedEventRecords.add(event);
			return false;
		} 
		else {
			//adding a new identifier entry to the Component to store identifiers that matched this component
			//this means all different classes including their interfaces
			Trace componentTrace = componentToTraceMap.get(component);
			if (componentTrace == null) {
				componentTrace = new Trace();
				componentToTraceMap.put(component, componentTrace);
			}
			componentTrace.addEvent(event);
			component.addComponentIdentifier(componentIdentifier);
			return true;
		}
	}

	private boolean addToResourceMap(ResourceRecord resourceEvent) {
		ResourceIdentifier resourceIdentifier = resourceEvent.getIdentifier();
		Resource resource = resourceIdentifierToResourceMap.get(resourceIdentifier);
		if (resource == null) {
			log.debug("resourceIdentifier of \"" + resourceEvent + "\" could not be mapped to a resource \n");
			undefinedResourceRecords.add(resourceEvent);
			return false;
		} else {
			Trace resourceTrace = resourceToTraceMap.get(resource);
			if (resourceTrace == null) {
				resourceTrace = new Trace();
				resourceToTraceMap.put(resource, resourceTrace);
			}
			resourceTrace.addResource(resourceEvent);
			resource.addResourceIdentifier(resourceIdentifier);
			return true;
		}
	}

	private void addToRequestMap(EventRecord event) {
		String requestID = event.getRequestIdentifier();
		Request request = idToRequestMap.get(requestID);
		if (request == null) {
			request = new Request(requestID);
			idToRequestMap.put(requestID, request);
		}
		request.addEvent(event);
	}



	public void printEventsPerRequest() {
		for (String requestID : idToRequestMap.keySet()) {
			System.out.println("Events for request " + requestID + " " + idToRequestMap.get(requestID).toString());
		}
	}

	public void printRequestsPerComponent() {
		for (Component component : componentToTraceMap.keySet()) {
			Trace componentTrace = componentToTraceMap.get(component);
			componentTrace.sortChronological();
			System.out.println("Events for component " + component.getName()
					+ " (" + component.getComponentIdentifiers() + ") "
					+ componentTrace);
		}

		System.out.println("Events which could not be mapped to components\t");
		for (EventRecord eventRecord : undefinedEventRecords) {
			System.out.println("\t" + eventRecord);
		}

	}

	public void printResources() {
		for (Resource resource : resourceToTraceMap.keySet()) {
			Trace resourceTrace = resourceToTraceMap.get(resource);
			resourceTrace.sortChronological();
			System.out.println("Events for resource " + resource.getName()
					+ "\n\t(" + resource.getResourceIdentifiers() + ") \n"
					+ resourceTrace + "\n");
		}

		System.out.println("Events which could not be mapped to resources\n");
		for (ResourceRecord record : undefinedResourceRecords) {
			System.out.println("\t" + record +"\n");
		}
	}

	public synchronized void setComponentDefinition(String componentName, String valueString) {
		if (componentIdentifierToComponentMap==null) {
			componentIdentifierToComponentMap = new HashMap<ComponentIdentifier, Component>();
		}
		Component component = new Component(componentName);


		//check existence of value string
		if (!valueString.equals(null)) {
			//divide string in several items
			String[] componentValues = valueString.split("&");

			if (componentValues.length==2) {
				String hostName = componentValues[0].trim();
				String componentDefinition = componentValues[1].trim();

				if (componentDefinition.startsWith("$")) {
					componentDefinition.replace("$", "").trim();
					componentIdentifierToComponentMap.put(ComponentIdentifier.createInterfaceComponentIdentifier(hostName,componentDefinition), component);
				} else {
					componentIdentifierToComponentMap.put(ComponentIdentifier.createClassComponentIdentifier(hostName,componentDefinition), component);
				}
			}

		}
	}

	public synchronized void setResourceDefinition(String resourceName, String valueString) {
		if (resourceIdentifierToResourceMap==null) {
			resourceIdentifierToResourceMap = new HashMap<ResourceIdentifier, Resource>();
		}
		Resource resource = new Resource(resourceName);
		//check existence of value string
		if (!valueString.equals(null)) {
			//divide string in several items
			String[] resourceValues = valueString.split("&");
			if (resourceValues.length>=1) {
				String hostName = resourceValues[0].trim();
				String resourceDefinition = null;
				if (resourceValues.length>=2) {
					resourceDefinition = resourceValues[1].trim();
				}
				resourceIdentifierToResourceMap.put(ResourceIdentifier.createResourceIdentifier(hostName,resourceDefinition),resource);
			}

		}

	}

}
